package engine

import (
	"fmt"
	"github.com/andybalholm/cascadia"
	"golang.org/x/net/html"
	"io"
	"reflect"
	"strings"
)

const attributePrefix = "data-"

type Engine struct {
	attributePrefix string
	errors          []error
	data            *map[string]any
	logNode         *html.Node
}

func (e *Engine) SetLogNode(node *html.Node) *Engine {
	e.logNode = node
	return e
}

func (e *Engine) SetAttributePrefix(prefix string) *Engine {
	e.attributePrefix = prefix
	return e
}

func (e *Engine) GetAttributePrefix() string {
	return e.attributePrefix
}

func (e *Engine) GetLogNode() *html.Node {
	return e.logNode
}

func New(data *map[string]any) *Engine {
	return &Engine{
		attributePrefix: attributePrefix,
		data:            data,
	}
}

func (e *Engine) HasErrors() bool {
	return len(e.errors) > 0
}

func (e *Engine) GetErrors() []error {
	return e.errors
}

func (e *Engine) ProcessNode(node *html.Node) *Engine {
	if node == nil {
		return e
	}

	if node.Type == html.ElementNode {
		e.ProcessElement(node)
	}
	return e
}

func functionExists(obj interface{}, funcName string) bool {
	m := reflect.ValueOf(obj).MethodByName(funcName)
	if m.IsValid() {
		return true
	}

	return false
}

func (e *Engine) callFunction(funcName string, node *html.Node, value string) *Engine {
	reflect.ValueOf(e).MethodByName(funcName).Call([]reflect.Value{reflect.ValueOf(node), reflect.ValueOf(value)})
	return e
}

func (e *Engine) ProcessElement(node *html.Node) *Engine {
	if node.Type != html.ElementNode {
		e.errors = append(e.errors, &UnsupportedTypeError{Message: "Unsupported element type", NodeType: node.Type})
		return e
	}

	for _, attr := range node.Attr {
		if attr.Namespace != "" ||
			len(attr.Key) < len(e.attributePrefix) ||
			attr.Key[:len(e.attributePrefix)] != e.attributePrefix {
			continue
		}

		fktName := attr.Key[len(e.attributePrefix):]
		fktName = buildFunctionName(fktName, true)

		fmt.Println(fktName)
		if !functionExists(e, fktName) {
			//			e.errors = append(e.errors, &UnsupportedFunctionError{Message: "Unsupported function", FunctionName: fktName})
			continue
		}

		e.callFunction(fktName, node, attr.Val)

	}

	return e

}

func isHtmlFragment(html string) bool {

	if strings.HasPrefix(html, "<!DOCTYPE html>") {
		return false
	}

	if strings.HasPrefix(html, "<html") {
		return false
	}

	return true
}

func (e *Engine) ProcessHtml(w io.Writer, r io.Reader) *Engine {

	var err error

	content, err := io.ReadAll(r)
	if err != nil {
		e.errors = append(e.errors, err)
		return e
	}

	stringContent := string(content)
	isFragment := isHtmlFragment(stringContent)

	var doc *html.Node

	selector := "body"

	if isFragment {
		c := strings.TrimSpace(stringContent)
		if strings.HasPrefix(c, "<tr") {
			selector = "table > tbody"
			c = "<table>" + c
		}

		if strings.HasSuffix(c, "</tr>") {
			c = c + "</table>"
		}

		if strings.HasPrefix(c, "<td") {
			selector = "table > tbody > tr"
			c = "<table><tr>" + c
		}

		if strings.HasSuffix(c, "</td>") {
			c = c + "</tr></table>"
		}

		stringContent = c
	}

	doc, err = html.Parse(strings.NewReader(stringContent))
	if err != nil {
		e.errors = append(e.errors, err)
		return e
	}

	e.ProcessNode(doc)

	if isFragment {

		sel, err := cascadia.Compile(selector)
		if err != nil {
			e.errors = append(e.errors, err)
			return e
		}

		body := sel.MatchFirst(doc)
		stringBuilder := &strings.Builder{}
		err = html.Render(stringBuilder, body)
		if err != nil {
			e.errors = append(e.errors, err)
		}

		result := stringBuilder.String()
		if selector == "body" {
			result = strings.TrimPrefix(result, "<body>")
			result = strings.TrimSuffix(result, "</body>")
		}

		if selector == "table > tbody" {
			result = strings.TrimPrefix(result, "<tbody>")
			result = strings.TrimSuffix(result, "</tbody>")
		}

		if selector == "table > tbody > tr" {
			result = strings.TrimPrefix(result, "<tr>")
			result = strings.TrimSuffix(result, "</tr>")
		}

		w.Write([]byte(result))

		return e

	}

	err = html.Render(w, doc)
	if err != nil {
		e.errors = append(e.errors, err)
	}
	return e

}

func (e *Engine) walkNodes(n *html.Node) *Engine {

	if n == nil {
		return e
	}

	for c := n.FirstChild; c != nil; c = c.NextSibling {
		e.ProcessNode(c)
		e.walkNodes(c)
	}

	return e
}

func (e *Engine) ProcessNodes(node *html.Node) *Engine {

	if node == nil {
		return e
	}

	e.ProcessNode(node)
	e.walkNodes(node)
	return e
}

func (e *Engine) ProcessAttributes(node *html.Node, value string) *Engine {

	value = strings.TrimSpace(value)
	values := strings.Split(value, ",")
	if len(values) == 0 {
		return e
	}

	for _, v := range values {
		v = strings.TrimSpace(v)
		if v == "" {
			continue
		}

		pos := strings.Index(v, " ")
		if pos == -1 {
			e.errors = append(e.errors, &InvalidAttributeError{
				Message:   "Invalid attribute",
				Attribute: v,
				Node:      node,
			})
			continue
		}

		key := v[:pos]
		instruction := v[pos+1:]

		cVal := e.getValue(instruction)
		node.Attr = append(node.Attr, html.Attribute{
			Key: key,
			Val: cVal,
		})

	}

	return e
}

func (e *Engine) getValue(instruction string) string {
	return ""
}

var uppercaseAcronym = map[string]string{
	"ID": "id",
}

// Converts a string to CamelCase
func buildFunctionName(s string, initCase bool) string {
	s = strings.TrimSpace(s)
	if s == "" {
		return s
	}
	if a, ok := uppercaseAcronym[s]; ok {
		s = a
	}

	n := strings.Builder{}
	n.Grow(len(s))
	capNext := initCase
	for i, v := range []byte(s) {
		vIsCap := v >= 'A' && v <= 'Z'
		vIsLow := v >= 'a' && v <= 'z'
		if capNext {
			if vIsLow {
				v += 'A'
				v -= 'a'
			}
		} else if i == 0 {
			if vIsCap {
				v += 'a'
				v -= 'A'
			}
		}
		if vIsCap || vIsLow {
			n.WriteByte(v)
			capNext = false
		} else if vIsNum := v >= '0' && v <= '9'; vIsNum {
			n.WriteByte(v)
			capNext = true
		} else {
			capNext = v == '_' || v == ' ' || v == '-' || v == '.'
		}
	}
	return "Process" + n.String()
}
