package engine

import (
	"github.com/stretchr/testify/assert"
	"golang.org/x/net/html"
	"strings"
	"testing"
)

func TestNewEngine(t *testing.T) {
	e := New(nil)
	assert.NotNil(t, e)
}

func TestProcessNode(t *testing.T) {

	testData := []struct {
		html      string
		hasErrors bool
		expected  string
		data      map[string]any
	}{
		{
			html:      `<div data-test="test"></div>`,
			hasErrors: false,
			expected:  `<div data-test="test"></div>`,
			data: map[string]any{
				"test": "test",
			},
		},
		{
			html:      `<div data-test="test"></div>`,
			hasErrors: false,
			expected:  `<div data-test="test"></div>`,
			data: map[string]any{
				"test": "test",
			},
		},
	}

	for _, td := range testData {
		e := New(&td.data)
		doc, _ := html.Parse(strings.NewReader(td.html))
		e.ProcessNode(doc)
		assert.Equal(t, td.hasErrors, e.HasErrors())
		//assert.Equal(t, td.expected, html.Render(doc))
	}

}

func TestProcessHtml(t *testing.T) {

	testData := []struct {
		html      string
		hasErrors bool
		expected  string
		data      map[string]any
	}{
		{
			html:      `<div data-test="test"></div>`,
			hasErrors: false,
			expected:  `<div data-test="test"></div>`,
			data:      map[string]any{},
		},
		{
			html:      `<div data-test="test"></div>`,
			hasErrors: false,
			expected:  `<div data-test="test"></div>`,
			data:      map[string]any{},
		},
		{
			html:      `<li>test1</li><li>test2</li>`,
			hasErrors: false,
			expected:  `<li>test1</li><li>test2</li>`,
			data:      map[string]any{},
		},
		{
			html:      `<tr><td>test1</td><td>test2</td></tr><tr><td>test1</td><td>test2</td></tr>`,
			hasErrors: false,
			expected:  `<tr><td>test1</td><td>test2</td></tr><tr><td>test1</td><td>test2</td></tr>`,
			data:      map[string]any{},
		},
		{
			html:      `<td>test1</td><td>test2</td>`,
			hasErrors: false,
			expected:  `<td>test1</td><td>test2</td>`,
			data:      map[string]any{},
		},
	}

	for _, td := range testData {
		e := New(&td.data)

		stringWriter := &strings.Builder{}
		e.ProcessHtml(stringWriter, strings.NewReader(td.html))
		processedDoc := stringWriter.String()
		assert.Equal(t, td.hasErrors, e.HasErrors())
		assert.Equal(t, td.expected, processedDoc)

	}

}

func BenchmarkParse(b *testing.B) {
	for i := 0; i < b.N; i++ {
		e := New(nil)
		doc, _ := html.Parse(strings.NewReader(`<div data-test="test"></div>`))
		e.ProcessNode(doc)
	}
}

func TestSetLogNode(t *testing.T) {
	e := New(nil)
	doc, _ := html.Parse(strings.NewReader(`<div data-test="test"></div>`))
	e.SetLogNode(doc.FirstChild)
	assert.True(t, doc.FirstChild.Type == html.ElementNode)
}

func TestSetLogNodeNil(t *testing.T) {
	e := New(nil)
	e.SetLogNode(nil)
}

func TestSetAttributePrefix(t *testing.T) {
	e := New(nil)
	e.SetAttributePrefix("test")
	assert.Equal(t, "test", e.attributePrefix)
}

func TestGetAttributePrefix(t *testing.T) {
	e := New(nil)
	e.SetAttributePrefix("test")
	assert.Equal(t, "test", e.GetAttributePrefix())
}

func TestProcessNodes(t *testing.T) {
	e := New(nil)
	doc, _ := html.Parse(strings.NewReader(`<div data-test="test">
<ul data-attributes="class static:hello">
<li>test1</li>
<li>test2</li>
<li><a href="test"></a></li>
</ul>

</div>`))
	e.ProcessNodes(doc)

	stringWriter := &strings.Builder{}
	html.Render(stringWriter, doc)
	assert.Equal(t, `<div data-test="test">`, stringWriter.String())

}

func TestProcessNodesNil(t *testing.T) {
	e := New(nil)
	e.ProcessNodes(nil)
}
